@@ -2,11 +2,12 @@
 
 from enum import Enum
 
+import pytz
 import six
-
-from brewtils.errors import RequestStatusTransitionError
+from brewtils.errors import ModelError, _deprecate
 
 __all__ = [
+    "BaseModel",
     "System",
     "Instance",
     "Command",
@@ -22,10 +23,14 @@ __all__ = [
     "Role",
     "RefreshToken",
     "Job",
+    "RequestFile",
     "RequestTemplate",
     "DateTrigger",
     "CronTrigger",
     "IntervalTrigger",
+    "FileTrigger",
+    "Garden",
+    "Operation",
 ]
 
 
@@ -36,29 +41,54 @@ class Events(Enum):
     BARTENDER_STOPPED = 4
     REQUEST_CREATED = 5
     REQUEST_STARTED = 6
+    REQUEST_UPDATED = 22
     REQUEST_COMPLETED = 7
+    REQUEST_CANCELED = 29
     INSTANCE_INITIALIZED = 8
     INSTANCE_STARTED = 9
+    INSTANCE_UPDATED = 23
     INSTANCE_STOPPED = 10
     SYSTEM_CREATED = 11
     SYSTEM_UPDATED = 12
     SYSTEM_REMOVED = 13
     QUEUE_CLEARED = 14
     ALL_QUEUES_CLEARED = 15
+    DB_CREATE = 16
+    DB_UPDATE = 17
+    DB_DELETE = 18
+    GARDEN_CREATED = 19
+    GARDEN_UPDATED = 20
+    GARDEN_REMOVED = 21
+    FILE_CREATED = 24
+    GARDEN_STARTED = 25
+    GARDEN_STOPPED = 26
+    GARDEN_UNREACHABLE = 27
+    GARDEN_ERROR = 28
+    GARDEN_NOT_CONFIGURED = 29
+    GARDEN_SYNC = 30
+    ENTRY_STARTED = 31
+    ENTRY_STOPPED = 32
+    JOB_CREATED = 33
+    JOB_DELETED = 34
+    JOB_PAUSED = 35
+    JOB_RESUMED = 36
+    PLUGIN_LOGGER_FILE_CHANGE = 37
 
 
-class Command(object):
+class BaseModel(object):
+    schema = None
 
+
+class Command(BaseModel):
     schema = "CommandSchema"
 
-    COMMAND_TYPES = ("ACTION", "INFO", "EPHEMERAL")
-    OUTPUT_TYPES = ("STRING", "JSON", "XML", "HTML")
+    COMMAND_TYPES = ("ACTION", "INFO", "EPHEMERAL", "ADMIN")
+    OUTPUT_TYPES = ("STRING", "JSON", "XML", "HTML", "JS", "CSS")
 
     def __init__(
         self,
         name=None,
         description=None,
-        id=None,
         parameters=None,
         command_type=None,
         output_type=None,
@@ -66,11 +96,10 @@ class Command(object):
         form=None,
         template=None,
         icon_name=None,
-        system=None,
+        hidden=False,
     ):
         self.name = name
         self.description = description
-        self.id = id
         self.parameters = parameters or []
         self.command_type = command_type
         self.output_type = output_type
@@ -78,7 +107,7 @@ class Command(object):
         self.form = form
         self.template = template
         self.icon_name = icon_name
-        self.system = system
+        self.hidden = hidden
 
     def __str__(self):
         return self.name
@@ -87,17 +116,39 @@ class Command(object):
         return "<Command: %s>" % self.name
 
     def parameter_keys(self):
-        """Convenience Method for returning all the keys of this command's parameters.
+        """Get a list of all Parameter keys
 
-        :return list_of_parameters:
+        Returns:
+            list[str]: A list containing each Parameter's key attribute
         """
         return [p.key for p in self.parameters]
 
-    def get_parameter_by_key(self, key):
-        """Given a Key, it will return the parameter (or None) with that key
+    def parameter_keys_by_type(self, desired_type):
+        """Get a list of all Parameter keys, filtered by Parameter type
 
-        :param key:
-        :return parameter:
+        Args:
+            desired_type (str): Parameter type
+
+        Returns:
+            list[str]: A list containing matching Parameters' key attribute
+        """
+        keys = []
+        for param in self.parameters:
+            key = param.keys_by_type(desired_type)
+            if key:
+                keys.append(key)
+        return keys
+
+    def get_parameter_by_key(self, key):
+        """Lookup a Parameter using a given key
+
+        Args:
+            key (str): The Parameter key to use
+
+        Returns:
+            Parameter (Optional): A Parameter with the given key
+
+            If a Parameter with the given key does not exist None will be returned.
         """
         for parameter in self.parameters:
             if parameter.key == key:
@@ -106,11 +157,13 @@ class Command(object):
         return None
 
     def has_different_parameters(self, parameters):
-        """Given a set of parameters, determines if the parameters provided differ from the
-        parameters already defined on this command.
+        """Determine if parameters differ from the current parameters
 
-        :param parameters:
-        :return boolean:
+        Args:
+            parameters (Sequence[Parameter]): Parameter collection for comparison
+
+        Returns:
+            bool: True if the given Parameters differ, False if they are identical
         """
         if len(parameters) != len(self.parameters):
             return True
@@ -126,8 +179,7 @@ class Command(object):
         return False
 
 
-class Instance(object):
-
+class Instance(BaseModel):
     schema = "InstanceSchema"
 
     INSTANCE_STATUSES = {
@@ -171,8 +223,7 @@ class Instance(object):
         return "<Instance: name=%s, status=%s>" % (self.name, self.status)
 
 
-class Choices(object):
-
+class Choices(BaseModel):
     schema = "ChoicesSchema"
 
     TYPES = ("static", "url", "command")
@@ -196,8 +247,7 @@ class Choices(object):
         )
 
 
-class Parameter(object):
-
+class Parameter(BaseModel):
     schema = "ParameterSchema"
 
     TYPES = (
@@ -209,6 +259,7 @@ class Parameter(object):
         "Dictionary",
         "Date",
         "DateTime",
+        "Bytes",
     )
     FORM_INPUT_TYPES = ("textarea",)
 
@@ -228,6 +279,7 @@ class Parameter(object):
         minimum=None,
         regex=None,
         form_input_type=None,
+        type_info=None,
     ):
         self.key = key
         self.type = type
@@ -243,6 +295,7 @@ class Parameter(object):
         self.minimum = minimum
         self.regex = regex
         self.form_input_type = form_input_type
+        self.type_info = type_info or {}
 
     def __str__(self):
         return self.key
@@ -254,6 +307,37 @@ class Parameter(object):
             self.description,
         )
 
+    def keys_by_type(self, desired_type):
+        """Gets all keys by the specified type.
+
+        Since parameters can be nested, this method will also return all keys of all
+        nested parameters. The return value is a possibly nested list, where the first
+        value of each list is going to be a string, while the next value is a list.
+
+        Args:
+            desired_type (str): Desired type
+
+        Returns:
+            An empty list if the type does not exist, otherwise it will be a list
+            containing at least one entry which is a string, each subsequent entry is a
+            nested list with the same structure.
+        """
+        keys = []
+        if self.type == desired_type:
+            keys.append(self.key)
+
+        if not self.parameters:
+            return keys
+
+        for param in self.parameters:
+            nested_keys = param.keys_by_type(desired_type)
+            if nested_keys:
+                if not keys:
+                    keys = [self.key]
+
+                keys.append(nested_keys)
+        return keys
+
     def is_different(self, other):
         if not type(other) is type(self):
             return True
@@ -261,6 +345,7 @@ class Parameter(object):
         fields_to_compare = [
             "key",
             "type",
+            "type_info",
             "multi",
             "optional",
             "default",
@@ -290,27 +375,62 @@ class Parameter(object):
         return False
 
 
-class RequestTemplate(object):
+class RequestFile(BaseModel):
+    schema = "RequestFileSchema"
 
+    def __init__(self, storage_type=None, filename=None):
+        self.storage_type = storage_type
+        self.filename = filename
+
+    def __str__(self):
+        return self.filename
+
+    def __repr__(self):
+        return "<RequestFile: filename=%s, storage_type=%s>" % (
+            self.filename,
+            self.storage_type,
+        )
+
+
+class RequestTemplate(BaseModel):
     schema = "RequestTemplateSchema"
 
+    TEMPLATE_FIELDS = [
+        "system",
+        "system_version",
+        "instance_name",
+        "namespace",
+        "command",
+        "command_type",
+        "parameters",
+        "comment",
+        "metadata",
+        "output_type",
+    ]
+
     def __init__(
         self,
         system=None,
         system_version=None,
         instance_name=None,
+        namespace=None,
         command=None,
+        command_type=None,
         parameters=None,
         comment=None,
         metadata=None,
+        output_type=None,
     ):
         self.system = system
         self.system_version = system_version
         self.instance_name = instance_name
+        self.namespace = namespace
         self.command = command
+        self.command_type = command_type
         self.parameters = parameters
         self.comment = comment
         self.metadata = metadata or {}
+        self.output_type = output_type
 
     def __str__(self):
         return self.command
@@ -318,25 +438,31 @@ class RequestTemplate(object):
     def __repr__(self):
         return (
             "<RequestTemplate: command=%s, system=%s, system_version=%s, "
-            "instance_name=%s>"
-            % (self.command, self.system, self.system_version, self.instance_name)
+            "instance_name=%s, namespace=%s>"
+            % (
+                self.command,
+                self.system,
+                self.system_version,
+                self.instance_name,
+                self.namespace,
+            )
         )
 
 
 class Request(RequestTemplate):
-
     schema = "RequestSchema"
 
     STATUS_LIST = ("CREATED", "RECEIVED", "IN_PROGRESS", "CANCELED", "SUCCESS", "ERROR")
     COMPLETED_STATUSES = ("CANCELED", "SUCCESS", "ERROR")
-    COMMAND_TYPES = ("ACTION", "INFO", "EPHEMERAL")
-    OUTPUT_TYPES = ("STRING", "JSON", "XML", "HTML")
+    COMMAND_TYPES = ("ACTION", "INFO", "EPHEMERAL", "ADMIN")
+    OUTPUT_TYPES = ("STRING", "JSON", "XML", "HTML", "JS", "CSS")
 
     def __init__(
         self,
         system=None,
         system_version=None,
         instance_name=None,
+        namespace=None,
         command=None,
         id=None,
         parent=None,
@@ -358,34 +484,53 @@ class Request(RequestTemplate):
             system=system,
             system_version=system_version,
             instance_name=instance_name,
+            namespace=namespace,
             command=command,
+            command_type=command_type,
             parameters=parameters,
             comment=comment,
             metadata=metadata,
+            output_type=output_type,
         )
         self.id = id
         self.parent = parent
         self.children = children
         self.output = output
-        self.output_type = output_type
         self._status = status
-        self.command_type = command_type
         self.created_at = created_at
         self.updated_at = updated_at
         self.error_class = error_class
         self.has_parent = has_parent
         self.requester = requester
 
+    @classmethod
+    def from_template(cls, template, **kwargs):
+        """Create a Request instance from a RequestTemplate
+
+        Args:
+            template: The RequestTemplate to use
+            **kwargs: Optional overrides to use in place of the template's attributes
+
+        Returns:
+            The new Request instance
+        """
+        request_params = {
+            k: kwargs.get(k, getattr(template, k))
+            for k in RequestTemplate.TEMPLATE_FIELDS
+        }
+        return Request(**request_params)
+
     def __repr__(self):
         return (
             "<Request: command=%s, status=%s, system=%s, system_version=%s, "
-            "instance_name=%s>"
+            "instance_name=%s, namespace=%s>"
             % (
                 self.command,
                 self.status,
                 self.system,
                 self.system_version,
                 self.instance_name,
+                self.namespace,
             )
         )
 
@@ -393,6 +538,10 @@ class Request(RequestTemplate):
     def status(self):
         return self._status
 
+    @status.setter
+    def status(self, value):
+        self._status = value
+
     @property
     def is_ephemeral(self):
         return self.command_type and self.command_type.upper() == "EPHEMERAL"
@@ -401,31 +550,8 @@ class Request(RequestTemplate):
     def is_json(self):
         return self.output_type and self.output_type.upper() == "JSON"
 
-    @status.setter
-    def status(self, value):
-        if self._status in self.COMPLETED_STATUSES:
-            raise RequestStatusTransitionError(
-                "Status for a request cannot be updated once it has been "
-                "completed. Current status: {0} Requested status: {1}".format(
-                    self.status, value
-                )
-            )
-
-        elif self._status == "IN_PROGRESS" and value not in self.COMPLETED_STATUSES + (
-            "IN_PROGRESS",
-        ):
-            raise RequestStatusTransitionError(
-                "A request cannot go from IN_PROGRESS to a non-completed "
-                "status. Completed statuses are {0}. You requested: {1}".format(
-                    self.COMPLETED_STATUSES, value
-                )
-            )
-
-        self._status = value
-
-
-class System(object):
 
+class System(BaseModel):
     schema = "SystemSchema"
 
     def __init__(
@@ -440,6 +566,8 @@ class System(object):
         icon_name=None,
         display_name=None,
         metadata=None,
+        namespace=None,
+        local=None,
     ):
         self.name = name
         self.description = description
@@ -451,12 +579,18 @@ class System(object):
         self.icon_name = icon_name
         self.display_name = display_name
         self.metadata = metadata or {}
+        self.namespace = namespace
+        self.local = local
 
     def __str__(self):
-        return "%s-%s" % (self.name, self.version)
+        return "%s:%s-%s" % (self.namespace, self.name, self.version)
 
     def __repr__(self):
-        return "<System: name=%s, version=%s>" % (self.name, self.version)
+        return "<System: name=%s, version=%s, namespace=%s>" % (
+            self.name,
+            self.version,
+            self.namespace,
+        )
 
     @property
     def instance_names(self):
@@ -465,27 +599,75 @@ class System(object):
     def has_instance(self, name):
         """Determine if an instance currently exists in the system
 
-        :param name: The name of the instance to search
-        :return: True if an instance with the given name exists for this system, False otherwise.
-        """
-        return True if self.get_instance(name) else False
+        Args:
+            name (str): The instance name
 
-    def get_instance(self, name):
+        Returns:
+            bool: True if an instance with the given name exists, False otherwise
+        """
+        return name in self.instance_names
+
+    def get_instance_by_name(self, name, raise_missing=False):
         """Get an instance that currently exists in the system
 
-        :param name: The name of the instance to search
-        :return: The instance with the given name exists for this system, None otherwise
+        Args:
+            name (str): The instance name
+            raise_missing (bool): If True, raise an exception if an Instance with the
+            given name is not found. If False, will return None in that case.
+
+        Returns:
+            Instance: The instance if it exists, None otherwise
+
+        Raises:
+            ModelError: Instance was not found and raise_missing=True
         """
         for instance in self.instances:
             if instance.name == name:
                 return instance
+
+        if raise_missing:
+            raise ModelError("Instance not found")
+
         return None
 
+    def get_instance_by_id(self, id, raise_missing=False):
+        """Get an instance that currently exists in the system
+
+        Args:
+            id (str): The instance id
+            raise_missing (bool): If True, raise an exception if an Instance with the
+            given id is not found. If False, will return None in that case.
+
+        Returns:
+            Instance: The instance if it exists, None otherwise
+
+        Raises:
+            ModelError: Instance was not found and raise_missing=True
+        """
+        for instance in self.instances:
+            if instance.id == id:
+                return instance
+
+        if raise_missing:
+            raise ModelError("Instance not found")
+
+        return None
+
+    def get_instance(self, name):
+        """DEPRECATED: Please use get_instance_by_name instead"""
+        _deprecate(
+            "Heads up! This method is deprecated, please use get_instance_by_name"
+        )
+        return self.get_instance_by_name(name)
+
     def get_command_by_name(self, command_name):
         """Retrieve a particular command from the system
 
-        :param command_name: Name of the command to retrieve
-        :return: The command object. None if the given command name does not exist in this system.
+        Args:
+            command_name (str): The command name
+
+        Returns:
+            Command: The command if it exists, None otherwise
         """
         for command in self.commands:
             if command.name == command_name:
@@ -496,8 +678,11 @@ class System(object):
     def has_different_commands(self, commands):
         """Check if a set of commands is different than the current commands
 
-        :param commands: The set commands to compare against the current set
-        :return: True if the sets are different, False if the sets are the same
+        Args:
+            commands (Sequence[Command]): Command collection for comparison
+
+        Returns:
+            bool: True if the given Commands differ, False if they are identical
         """
         if len(commands) != len(self.commands):
             return True
@@ -514,8 +699,7 @@ class System(object):
         return False
 
 
-class PatchOperation(object):
-
+class PatchOperation(BaseModel):
     schema = "PatchSchema"
 
     def __init__(self, operation=None, path=None, value=None):
@@ -534,8 +718,7 @@ class PatchOperation(object):
         )
 
 
-class LoggingConfig(object):
-
+class LoggingConfig(BaseModel):
     schema = "LoggingConfigSchema"
 
     LEVELS = ("DEBUG", "INFO", "WARN", "ERROR")
@@ -571,12 +754,16 @@ class LoggingConfig(object):
     def get_plugin_log_config(self, **kwargs):
         """Get a specific plugin logging configuration.
 
-        It is possible for different systems to have different logging configurations. This method
-        will create the correct plugin logging configuration and return it. If a specific logger
-        is not found for a system, then the current logging configuration will be returned.
+        It is possible for different systems to have different logging configurations.
+        This method will create the correct plugin logging configuration and return it.
+        If a specific logger is not found for a system, then the current logging
+        configuration will be returned.
 
-        :param kwargs: Identifying information for a system (i.e. system_name)
-        :return:
+        Keyword Args:
+            Identifying information for a system (i.e. system_name)
+
+        Returns:
+            The logging configuration for this system
         """
         system_name = kwargs.pop("system_name", None)
         specific_logger = self._loggers.get(system_name, {})
@@ -636,33 +823,53 @@ class LoggingConfig(object):
         )
 
 
-class Event(object):
-
+class Event(BaseModel):
     schema = "EventSchema"
 
     def __init__(
-        self, name=None, payload=None, error=None, metadata=None, timestamp=None
+        self,
+        name=None,
+        namespace=None,
+        garden=None,
+        metadata=None,
+        timestamp=None,
+        payload_type=None,
+        payload=None,
+        error=None,
+        error_message=None,
     ):
         self.name = name
-        self.payload = payload
-        self.error = error
+        self.namespace = namespace
+        self.garden = garden
         self.metadata = metadata or {}
         self.timestamp = timestamp
+        self.payload_type = payload_type
+        self.payload = payload
+        self.error = error
+        self.error_message = error_message
 
     def __str__(self):
-        return "%s: %s, %s" % (self.name, self.payload, self.metadata)
+        return "%s: %s" % (self.namespace, self.name)
 
     def __repr__(self):
-        return "<Event: name=%s, error=%s, payload=%s, metadata=%s>" % (
-            self.name,
-            self.error,
-            self.payload,
-            self.metadata,
+        return (
+            "<Event: namespace=%s, garden=%s, name=%s, timestamp=%s, error=%s, "
+            "error_message=%s, metadata=%s, payload_type=%s, payload=%r>"
+            % (
+                self.namespace,
+                self.garden,
+                self.name,
+                self.timestamp,
+                self.error,
+                self.error_message,
+                self.metadata,
+                self.payload_type,
+                self.payload,
+            )
         )
 
 
-class Queue(object):
-
+class Queue(BaseModel):
     schema = "QueueSchema"
 
     def __init__(
@@ -690,8 +897,7 @@ class Queue(object):
         return "<Queue: name=%s, size=%s>" % (self.name, self.size)
 
 
-class Principal(object):
-
+class Principal(BaseModel):
     schema = "PrincipalSchema"
 
     def __init__(
@@ -721,32 +927,23 @@ class Principal(object):
         )
 
 
-class Role(object):
-
+class Role(BaseModel):
     schema = "RoleSchema"
 
-    def __init__(
-        self, id=None, name=None, description=None, roles=None, permissions=None
-    ):
+    def __init__(self, id=None, name=None, description=None, permissions=None):
         self.id = id
         self.name = name
         self.description = description
-        self.roles = roles
         self.permissions = permissions
 
     def __str__(self):
         return "%s" % self.name
 
     def __repr__(self):
-        return "<Role: name=%s, roles=%s, permissions=%s>" % (
-            self.name,
-            self.roles,
-            self.permissions,
-        )
+        return "<Role: name=%s, permissions=%s>" % (self.name, self.permissions)
 
 
-class RefreshToken(object):
-
+class RefreshToken(BaseModel):
     schema = "RefreshTokenSchema"
 
     def __init__(self, id=None, issued=None, expires=None, payload=None):
@@ -766,9 +963,8 @@ class RefreshToken(object):
         )
 
 
-class Job(object):
-
-    TRIGGER_TYPES = {"interval", "date", "cron"}
+class Job(BaseModel):
+    TRIGGER_TYPES = {"interval", "date", "cron", "file"}
     STATUS_TYPES = {"RUNNING", "PAUSED"}
     schema = "JobSchema"
 
@@ -807,7 +1003,7 @@ class Job(object):
         return "<Job: name=%s, id=%s>" % (self.name, self.id)
 
 
-class DateTrigger(object):
+class DateTrigger(BaseModel):
     schema = "DateTriggerSchema"
 
     def __init__(self, run_date=None, timezone=None):
@@ -820,8 +1016,18 @@ class DateTrigger(object):
     def __repr__(self):
         return "<DateTrigger: run_date=%s>" % self.run_date
 
+    @property
+    def scheduler_attributes(self):
+        return ["run_date", "timezone"]
 
-class IntervalTrigger(object):
+    @property
+    def scheduler_kwargs(self):
+        tz = pytz.timezone(self.timezone)
+
+        return {"timezone": tz, "run_date": tz.localize(self.run_date)}
+
+
+class IntervalTrigger(BaseModel):
     schema = "IntervalTriggerSchema"
 
     def __init__(
@@ -858,8 +1064,38 @@ class IntervalTrigger(object):
             % (self.weeks, self.days, self.hours, self.minutes, self.seconds)
         )
 
+    @property
+    def scheduler_attributes(self):
+        return [
+            "weeks",
+            "days",
+            "hours",
+            "minutes",
+            "seconds",
+            "start_date",
+            "end_date",
+            "timezone",
+            "jitter",
+            "reschedule_on_finish",
+        ]
 
-class CronTrigger(object):
+    @property
+    def scheduler_kwargs(self):
+        tz = pytz.timezone(self.timezone)
+
+        kwargs = {key: getattr(self, key) for key in self.scheduler_attributes}
+        kwargs.update(
+            {
+                "timezone": tz,
+                "start_date": tz.localize(self.start_date) if self.start_date else None,
+                "end_date": tz.localize(self.end_date) if self.end_date else None,
+            }
+        )
+
+        return kwargs
+
+
+class CronTrigger(BaseModel):
     schema = "CronTriggerSchema"
 
     def __init__(
@@ -901,3 +1137,162 @@ class CronTrigger(object):
             self.month,
             self.day,
         )
+
+    @property
+    def scheduler_attributes(self):
+        return [
+            "year",
+            "month",
+            "day",
+            "week",
+            "day_of_week",
+            "hour",
+            "minute",
+            "second",
+            "start_date",
+            "end_date",
+            "timezone",
+            "jitter",
+        ]
+
+    @property
+    def scheduler_kwargs(self):
+        tz = pytz.timezone(self.timezone)
+
+        kwargs = {key: getattr(self, key) for key in self.scheduler_attributes}
+        kwargs.update(
+            {
+                "timezone": tz,
+                "start_date": tz.localize(self.start_date) if self.start_date else None,
+                "end_date": tz.localize(self.start_date) if self.start_date else None,
+            }
+        )
+
+        return kwargs
+
+
+class FileTrigger(BaseModel):
+    schema = "FileTriggerSchema"
+
+    def __init__(
+        self,
+        pattern=None,
+        path=None,
+        recursive=None,
+        callbacks=None,
+    ):
+        self.pattern = pattern
+        self.path = path
+        self.recursive = recursive
+        self.callbacks = callbacks
+
+    def __str__(self):
+        return repr(self)
+
+    def __repr__(self):
+        return "<FileTrigger: %s %s %s %s>" % (
+            self.pattern,
+            self.path,
+            self.recursive,
+            self.callbacks,
+        )
+
+    @property
+    def scheduler_attributes(self):
+        return ["pattern", "path", "recursive", "callbacks"]
+
+    @property
+    def scheduler_kwargs(self):
+        kwargs = {key: getattr(self, key) for key in self.scheduler_attributes}
+        kwargs.update(
+            {
+                "pattern": self.pattern,
+                "path": self.path,
+                "recursive": self.recursive,
+                "callbacks": self.callbacks,
+            }
+        )
+
+        return kwargs
+
+
+class Garden(BaseModel):
+    schema = "GardenSchema"
+
+    GARDEN_STATUSES = {
+        "INITIALIZING",
+        "RUNNING",
+        "BLOCKED",
+        "STOPPED",
+        "NOT_CONFIGURED",
+        "UNREACHABLE",
+        "ERROR",
+        "UNKNOWN",
+    }
+
+    def __init__(
+        self,
+        id=None,
+        name=None,
+        status=None,
+        status_info=None,
+        namespaces=None,
+        systems=None,
+        connection_type=None,
+        connection_params=None,
+    ):
+        self.id = id
+        self.name = name
+        self.status = status.upper() if status else None
+        self.status_info = status_info or {}
+        self.namespaces = namespaces or []
+        self.systems = systems or []
+
+        self.connection_type = connection_type
+        self.connection_params = connection_params
+
+    def __str__(self):
+        return "%s" % self.name
+
+    def __repr__(self):
+        return "<Garden: garden_name=%s, status=%s>" % (self.name, self.status)
+
+
+class Operation(BaseModel):
+    schema = "OperationSchema"
+
+    def __init__(
+        self,
+        model=None,
+        model_type=None,
+        args=None,
+        kwargs=None,
+        target_garden_name=None,
+        source_garden_name=None,
+        operation_type=None,
+    ):
+        self.model = model
+        self.model_type = model_type
+        self.args = args or []
+        self.kwargs = kwargs or {}
+        self.target_garden_name = target_garden_name
+        self.source_garden_name = source_garden_name
+        self.operation_type = operation_type
+
+    def __str__(self):
+        return "%s" % self.operation_type
+
+    def __repr__(self):
+        return (
+            "<Operation: operation_type=%s, source_garden_name=%s, "
+            "target_garden_name=%s, model_type=%s, model=%s, args=%s, kwargs=%s>"
+            % (
+                self.operation_type,
+                self.source_garden_name,
+                self.target_garden_name,
+                self.model_type,
+                self.model,
+                self.args,
+                self.kwargs,
+            )
+        )
